/***************************************************************************
 *            tranches.cpp
 *
 *  Copyright  2005  Florent Berthaut
 *  florentberthaut@no-log.org
 ****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "tranches.h"
#include <math.h>
#include "traParser.h"
#include <sstream>
using namespace std;

tranches :: tranches(int argc,char** argv):jack(NULL),midi(NULL)
{
	//default slicesArray
	std::vector<int> vec;
	slicesArrays.push_back(vec);
	slicesArrays[0].push_back(-32);
	slicesArrays[0].push_back(-16);
	slicesArrays[0].push_back(-4);
	slicesArrays[0].push_back(-2);
	slicesArrays[0].push_back(1);
	slicesArrays[0].push_back(2);
	slicesArrays[0].push_back(3);
	slicesArrays[0].push_back(4);
	currentSlicesArray=0;

	//default midi controls
	for(int i=0;i<NB_MIDI_CONTROLS;++i)
	{
		midiControls[i][0]=-1;
		midiControls[i][1]=-1;
	}

	fileName="";

	#ifdef WITH_GUI 
		//start the ui
		execWin = new execWindow(300,350,"Tranches",this);	
		execWin->size_range(350,400,0,0,0,0);
		execWin->resizable(execWin);
		execWin->show();
	#endif
	
	//jackProcess
	jack = new jackProcess(this);
	if(jack->init())
	{
		showMessage(true,"Error Initialiasing Jack Client");
	}

	//midiProcess
	midi = new midiProcess(this,jack->getClient(),MIDI_PRIORITY);	
	showMessage(false,"Starting Midi");
	if(midi->midiInit())
	{
		showMessage(true,"Error Starting Midi ");
	}
	midi->startThread();

	//if a file name was specified
	if(argc==2)
	{
		load(argv[1]);
	}


	#ifdef WITH_LASH
		//lashClient
		lashC = new lashClient(this,argc,argv); 
		if(lashC->init())
		{
			showMessage(false,"Error initializing Lash, disabling lash");
		}
		else
		{
			//start lash client
			showMessage(false,"Starting LASH Client");
			lashC->start();
			sleep(2);
		}
	#endif

	//start 
	start();

	//if everything's ok  
	showMessage(false,"Let's slice!!");

	#ifndef WITH_GUI
		while(loop)
		{
			showMessage(false,"Menu :");
			showMessage(false,"1 : Open File");
			showMessage(false,"2 : Reset/Reload File");
			showMessage(false,"3 : Quit");
			int choice;
			while(!(cin>>choice) || choice<1 || choice>3)
			{
				if ( cin.fail() )
				{ 
				    cin.clear(); 
				    cin.ignore( numeric_limits<streamsize>::max(), '\n' );
				}		
				showMessage(false,"Incorrect Choice");
				showMessage(false,"Menu :");
				showMessage(false,"1 : Open File");
				showMessage(false,"2 : Reset/Reload File");
				showMessage(false,"3 : Quit");
			}
			if(choice==1)
			{
				std::string f;
				showMessage(false,"Enter the file name:");
				cin>>f;
				cout<<f;
				size_t size = f.size() + 1;
				char * cf = new char[ size ];
				strncpy( cf, f.c_str(), size );
				load(cf);
				start();
				delete [] cf;
			}
			else if(choice==2)
			{
				size_t size = fileName.size() + 1;
				char * cf = new char[ size ];
				strncpy( cf, fileName.c_str(), size );
				load(cf);
				start();
				delete [] cf;
			}
			else if(choice==3)
			{
				stop();
				loop=false;
			}
		}
	#endif
}


tranches::~tranches()
{
}

void tranches::showMessage(bool e,std::string mess)
{
	#ifdef WITH_GUI
		execWin->showMessage(e,mess);
	#else
		cout<<"<Tranches> "<<mess<<endl;
	#endif
}

std::string tranches::getFileName()
{
	size_t found = fileName.find_last_of("/\\");	
	return fileName.substr(found+1);
}

std::string tranches::getCompleteFileName()
{
	return fileName;
}

int tranches::load(char* nomfic)
{
	//delete the outputs and inputs
	jackInputs.clear();
	jackOutputs.clear();

	fileName=nomfic;

 	// File Open
	traParser parse(this);
	if(parse.createTree(nomfic))
		return -1;
	showMessage(false,"Loading "+std::string(nomfic));
	if(parse.parseTree())
	{
		showMessage(true,"Error Parsing "+getFileName());
		return -1;
	}
	showMessage(false,getFileName()+" loaded");
		
	return 0;
}

int tranches::start()
{
	//test the input number and the output number
	if(jackInputs.size()==0)
	{
		jackInputs.push_back("input");
		inputsMidiCCs.push_back(-1);
		inputsMidiChannels.push_back(-1);
	}
	if(jackOutputs.size()==0)
	{
		jackOutputs.push_back("output");
	}

	if(jack->start())
	{
		showMessage(true,"Error starting jack client. Did you start Jack ?");
	}

	//(re)init lash
	#ifdef WITH_LASH
		sleep(2);
		lashC->setAlsa(midi->getAlsaID());
		lashC->setJack();
	#endif

	//(re)init the gui
	#ifdef WITH_GUI
		execWin->setTitle("Tranches : "+getFileName());
		execWin->init();
	#endif

	return 0;
}



void tranches::save(char* fileN)
{
	fileName=fileN;
	traParser tra(this);
	if(!tra.saveToFile(fileN))
	{
		#ifdef WITH_GUI
			execWin->setTitle("Tranches : "+getFileName());
		#endif
	}
}


void tranches::stop()
{
	//stops the midi thread
	if(midi)
	{
		midi->stopThread();
		delete midi;
		midi=NULL;
	}

	//stops the jack client
	if(jack)
	{
		delete jack;	
		jack=NULL;
	}

	//delete the outputs and inputs
	jackInputs.clear();
	jackOutputs.clear();

	loop=false;

	#ifdef WITH_GUI
		//reset the gui
		execWin->reset();
	#endif
}

int tranches::getOutputNb()
{
	return jackOutputs.size(); 
}

void tranches::addOutput(std::string name)
{
	jackOutputs.push_back(name);
}

void tranches::addCreateOutput(std::string name)
{
	jackOutputs.push_back(name);
	jack->addPort(1,name);
}

void tranches::removeOutput(int o)
{
	jackOutputs.erase(jackOutputs.begin()+o);
	jack->removePort(1,o);
}

int tranches::getOutputByName(std::string name)
{
	for(int i=0;i<jackOutputs.size();++i)
	{
		if(jackOutputs[i] == name)
		{
			return i;
		}	
	}
	return 0;
}

std::string tranches::getOutputName(int num)
{
	return jackOutputs[num];
}

void tranches::setOutputName(int num,std::string name)
{
	jackOutputs[num]=name;
	jack->renamePort(1,num,name);
}

int tranches::getInputNb()
{
	return jackInputs.size(); 
}


void tranches::addInput(std::string name,short chan,short cc)
{
	jackInputs.push_back(name);
	inputsMidiChannels.push_back(chan);
	inputsMidiCCs.push_back(cc);
}

void tranches::addCreateInput(std::string name,short chan,short cc)
{
	jackInputs.push_back(name);
	inputsMidiChannels.push_back(chan);
	inputsMidiCCs.push_back(cc);
	jack->addPort(0,name);
}

void tranches::removeInput(int i)
{
	jackInputs.erase(jackInputs.begin()+i);
	inputsMidiChannels.erase(inputsMidiChannels.begin()+i);
	inputsMidiCCs.erase(inputsMidiCCs.begin()+i);
	jack->removePort(0,i);
}

int tranches::getInputByName(std::string name)
{
	for(int i=0;i<jackInputs.size();++i)
	{
		if(jackInputs[i] == name)
		{
			return i;
		}	
	}
	return 0;
}

std::string tranches::getInputName(int num)
{
	return jackInputs[num];
}

void tranches::setInputName(int num,std::string name)
{
	jackInputs[num]=name;
	jack->renamePort(0,num,name);
}

void tranches::startMidiLearn()
{
	midi->startMidiLearn();
}

void tranches::stopMidiLearn()
{
	midi->stopMidiLearn();
}

void tranches::processMidiLearn(int ch,int ccn)
{
	#ifdef WITH_GUI
		execWin->learnMidi(ch,ccn);
	#endif
}

void tranches::setMidiControl(int control,int ccc,short midi)
{
	//if normal midi control
	if(control<NB_MIDI_CONTROLS)
		midiControls[control][ccc]=midi;
	//if midi input control
	else if(control-NB_MIDI_CONTROLS<inputsMidiChannels.size())
		if(!ccc)//channel
			inputsMidiChannels[control-NB_MIDI_CONTROLS]=midi;
		else//cc
			inputsMidiCCs[control-NB_MIDI_CONTROLS]=midi;
}

short tranches::getMidiControl(int control ,int ccc)
{
	//if normal midi control
	if(control<NB_MIDI_CONTROLS)
		return midiControls[control][ccc];
	//if midi input control
	else if(control-NB_MIDI_CONTROLS<inputsMidiChannels.size())
		if(!ccc)//channel
			return inputsMidiChannels[control-NB_MIDI_CONTROLS];
		else//cc
			return inputsMidiCCs[control-NB_MIDI_CONTROLS];
	else return -1;
	
}
	
int tranches::processMidiCC(unsigned short chan,unsigned short cc,unsigned short val)
{
	int controle=-1;
	for(int i=0;i<NB_MIDI_CONTROLS;++i)
		if(midiControls[i][0]==chan && midiControls[i][1]==cc)
		{
			controle=i;
			break;	
		}

	unsigned short res=-1;
	
/*
	switch(controle)
	{
		case 0://Input
			setInput(val);
			res=val;
			break;
		case 1://Output
			setOutput(val);
			res=val;
			break;
		case 2://slice
			setSlice(val);
			res=val;
			break;	
		case 3://swap
			//repeat(val);
			res=val;
			break;
		case 4://swap slice 1
			//repeat(val);
			res=val;
			break;
		case 5://swap slice 2
			//repeat(val);
			res=val;
			break;
		case 6://repeat
			repeat(val);
			res=val;
			break;
		case 7://reverse
			reverse(val);
			#ifdef WITH_GUI
				execWin->toggleExecReverse(val);
			#endif
			res=val;
			break;
		case 8://granulize
			//repeat(val);
			res=val;
			break;
		case 9://grain size
			//reverse(val);
			#ifdef WITH_GUI
				execWin->toggleExecReverse(val);
			#endif
			res=val;
			break;
		default://not a control
			break;
	}
	*/

	switch(controle)
	{
		case 0://Input
			setInput(val,false);
			res=val;
			break;
		case 1://Output
			setOutput(val);
			res=val;
			break;
		case 2://slice
			setSlice(val,false);
			res=val;
			break;	
		case 3://repeat
			repeat(val,false);
			res=val;
			break;
		case 4://reverse
			reverse(val);
			res=val;
			break;
		default://not a control
			break;
	}

	return res;	
}

void tranches::repeat(int val,bool fromGui)
{
	if(val==0)
		jack->stopSlicing();
	else
		jack->startSlicing();

	#ifdef WITH_GUI
		if(!fromGui)
			execWin->toggleExecRepeat(val);
	#endif
}

void tranches::reverse(int val)
{
	jack->toggleReverse(val);
	#ifdef WITH_GUI
		execWin->toggleExecReverse(val);
	#endif
}

void tranches::setSlice(int val,bool fromGui)
{
	//we compute the number of beats for the repetition according to the midi value
	//double nbBeats = (val<=5)? 1.0/pow(2.0,5-val) : val-4;
		
	//we get the corresponding slice size int the current slicesArray
	if(val>=0 && val<slicesArrays[currentSlicesArray].size())
	{
		//if slice < 0 -> (1 / slice) beats
		if(slicesArrays[currentSlicesArray][val]<0)
			jack->setSliceSize(-1.0/(double)(slicesArrays[currentSlicesArray][val]));	
		//if slice > 0 -> (slice) beats
		else if(slicesArrays[currentSlicesArray][val]>0)
			jack->setSliceSize((double)(slicesArrays[currentSlicesArray][val]));	
		#ifdef WITH_GUI
			//update the gui
			if(!fromGui)
				execWin->changeSlice(val,slicesArrays[currentSlicesArray][val]);
		#endif
	}
}

int tranches::getSlice(int val)
{
	//we get the corresponding slice size int the current slicesArray
	if(val>=0 && val<slicesArrays[currentSlicesArray].size())
	{
		return slicesArrays[currentSlicesArray][val];
	}
	else
		return 1;
}

void tranches::setInput(int val,bool fromGui)
{
	if(val>=jackInputs.size() || val<0) // bigger than the number of inputs -> select all inputs
	{
		jack->setSliceInput(-1);	
		#ifdef WITH_GUI
			if(!fromGui)
				execWin->changeInput(0);
		#endif
	}
	else	//just select the corresponding input
	{	
		jack->setSliceInput(val);
		#ifdef WITH_GUI
			if(!fromGui)
				execWin->changeInput(val+1);
		#endif
	}
}

void tranches::setOutput(int val)
{
	int out = (val>=jackOutputs.size())?jackOutputs.size()-1:val;
	jack->setSliceOutput(out);
	#ifdef WITH_GUI
		execWin->changeOutput(out);
	#endif
}


